# -*- test-case-name: wokkel.test.test_adhoc -*-
#
# Copyright (c) 2003-2008 Ralph Meijer
# See LICENSE for details.

"""
Ad-Hoc

Support for Ad-Hoc commands as described in
U{XEP-0004<http://www.xmpp.org/extensions/xep-0050.html>}.
"""
import types
from twisted.words.protocols.jabber.jid import JID
from twisted.words.xish import domish
from twisted.words.protocols.jabber import xmlstream
from wokkel import data_form

NS_ADHOC = 'http://jabber.org/protocol/commands'

IQ_GET_COMMAND = "/iq[@type='get']/command"
IQ_SET_COMMAND = "/iq[@type='set']/command"


class Error(Exception):
    """
    Data Forms error.
    """

class BadRequest(Error):
    """ bad-request
    """

class BadLocale(Error):
    """ bad-locate
    """

class Actions(object):
    """Representing a Actions in an Ad-Hoc command.
    See: http://xmpp.org/extensions/xep-0050.html#desc-actions

    @ivar execute: The action to execute.
    @type execute: L{str}

    @ivar allowed: An allowed list of execute actions.
    @type allowed: L{list} or L{str}
        
    """

    _ALLOWED = ['execute', 'cancel', 'prev', 'next', 'complete'] 
    
    def __init__(self, execute=None, allowed=[]):
        """
        Create an ad-hoc actions for stages in commands.

        """
        self.execute = execute
        self.allowed = allowed

    def toElement(self):
        """
        Return the DOM representation of this command.

        @rtype L{domish.Element}.

        """

        actions = domish.Element((None, 'actions'))
        if self.execute is not None:
            actions['execute'] = self.execute

        for a in self.allowed:
            if a in self._ALLOWED:
                actions.addElement(a)

        return actions

    @staticmethod
    def fromElement(element):
        if ((element.uri, element.name)) != ((None, 'actions')):
            raise Error("Element provided is incorrect.")
        actions = Actions()

        actions.execute = element.getAttribute('execute')
        
        for child in element.elements():
            if child.name in actions._ALLOWED:
                actions.allowed.append(child.name)
        return actions

class Note(object):
    """Representing a Note in an Ad-Hoc command.

    See: http://xmpp.org/extensions/xep-0050.html#desc-note

    @ivar note: The message for the note.
    @type note: L{str}        

    @ivar type: The type of note.
    @type type: L{str}

    """

    def __init__(self, note, type='info'):
        """
        Create a note for Ad-Hoc commands.

        """
        self.note = note
        self.type = type

    def toElement(self):
        """
        Return the DOM representation of this command.

        @rtype L{domish.Element}.

        """
        n = domish.Element((NS_ADHOC, 'note'))
        
        n['type'] = self.type

        n.addChild(self.note)

        return n

    @staticmethod
    def fromElement(element):
        """Create a L{Note} from a L{domish.Element}    
        """
        if (element.uri, element.name) != ((None, 'note')):
            raise Error("Element provided is not a Command Note")
        cdata = ''
        for child in element.children:
            if isinstance(child, types.StringTypes):
                cdata += child
        return Note(cdata, type=element.getAttribute('type','info'))



class Command(object):
    """A ad hoc command class representing an ad-hoc command.
    
    See: http://xmpp.org/extensions/xep-0050.html

    @ivar command: The command to take action on.
    @type command: L{str}
    
    @ivar action: The acction to take on the command.
    @type action: L{str}

    @ivar sessionid: The session id of this command session.
    @type sessionid: L{unicode}
    
    @ivar status: The status of this command.
    @type status: L{str}
        
    @ivar actions: The actions that are allowed or have been executed in the command.
    @type actions: L{Actions}

    @ivar note: The note attached to the command.
    @type note: L{Note}
    
    @ivar form: The data form used in this command.
    @type form: L{data_form.Form}

    """

    def __init__(self, command, action=None, sessionid=None, status=None, actions=None, note=None, form=None):
        """
        Create the command.
               
        """

        self.command = command
        self.action  = action


        self.sessionid = sessionid
        self.status    = status
        
        self.actions = actions
        self.note    = note
        self.form    = form

    def toElement(self):
        """
        Return the DOM representation of this command.

        @rtype L{domish.Element}.

        """
        c = domish.Element((NS_ADHOC, 'command'))
        
        c['node'] = self.command
        if self.action is not None:
            c['action'] = self.action

        if self.sessionid is not None:
            c['sessionid'] = self.sessionid

        if self.status is not None:
            c['status'] = self.status


        if self.note is not None:
            c.addChild(self.note.toElement())

        if self.actions is not None:
            c.addChild(self.actions.toElement())
        
        if self.form is not None:
            c.addChild(self.form.toElement())

        return c

    @staticmethod
    def fromElement(element):
        """Create a L{Command} from a L{domish.Element}

        @param element: The element representing the command.
        @type  element: L{domish.Element}

        """

        if (element.uri, element.name) != ((NS_ADHOC, 'command')):
            raise Error("Element provided is not a Command ")
        if not element.hasAttribute('node'):
            raise Error("Element provided is not a Command ")
        
        c = Command(element.getAttribute('node'))

        for attr in element.attributes:
            if attr != 'node':
                setattr(c, attr, element[attr])

        note = getattr(element, 'note', None)
        if note is not None:
            c.note = Note.fromElement(note)

        actions = getattr(element, 'actions', None)
        if actions is not None:
            c.actions = Actions.fromElement(actions)

        x = getattr(element, 'x', None)
        if x and x.uri == data_form.NS_X_DATA:
            c.form = data_form.Form.fromElement(x)

        return c


class CommandRequest(xmlstream.IQ):
    """ An Ad-Hoc Command Request.
    
    @ivar method: Type attribute of the IQ request. Either C{'set'} or C{'get'}
    @type method: C{str}
    
    @ivar command: Command we want to execute.
    @type command: L{Command}
    
    """

    
    def __init__(self, xs, command, method='set'):
        xmlstream.IQ.__init__(self, xs, method)
        self.addChild(command.toElement())
