"""
Asterisk Management Interface related mechanisms
Extends starpy with custom functionality
Author: DanB [dan@itsyscom.com]
License: GPLv3
"""
from twisted.internet import protocol, defer, reactor
from twisted.python import log
from mbitrans.manager import MBITAMIProtocol
from mbitrans.settingsdb import SettingsDB
from mbitrans.extensdb import ExtensDB
from mbitrans.callsdb import CallsDB
from mbitrans.parkdb import ParkDB
from mbitrans.confdb import ConfDB
from mbitrans.mailboxdb import MailboxDB
from mbitrans.queuedb import QueueDB
from mbitrans import libmbit

from time import time
import re


EXTEN_UNAVAILABLE = "4"
EXTENSION_STATES = {
	'-1': 'NOT_FOUND',
	'0': 'AVAILABLE',
	'1': 'ON THE PHONE',
	'2': 'BUSY',
	'4': 'UNAVAILABLE',
	'8': 'RINGING',
	'16': 'ON HOLD'
	}


class MBitAMIFactory( protocol.ClientFactory ):
	"""A factory for AMI protocols
	"""
	protocol = MBITAMIProtocol
	connections = []
	
	def __init__(self, settings):
		self.settings = settings
		self.username = self.settings['ami_user']
		self.secret = self.settings['ami_pwd']
		self.loginDefer = defer.Deferred() #Protocol instance will call this function on login success or failure
		self.loginDefer.addCallbacks( self.loginSuccess, log.err )

	def getConn( self ):
		"""
		@returns: first available connection to asterisk instance
		"""
		if not self.connections:
			raise Exception("No Asterisk connection!")
		return self.connections[0]

	def clientConnectionFailed( self, connector, reason ):
		"""Connection failed, report to our callers"""
		self.loginDefer.errback( reason )

	def loginSuccess( self, conn ):
		"""Login has succeeded, cache connection to directly call it's methods
		ToDo: 
			* Uncache on connection lost
			* Register event handlers here
		"""
		self.connections.append( conn )
		self.onConnection()
		
	def onConnection( self ):
		self.getConn().registerEvent('ExtensionStatus', self.eventExtensionStatus)
		self.getConn().registerEvent('Newchannel', self.eventCallMonitor)
		self.getConn().registerEvent('Dial', self.eventCallMonitor)
		self.getConn().registerEvent('Newstate', self.eventCallMonitor)
		self.getConn().registerEvent('Hangup', self.eventCallMonitor)
		self.getConn().registerEvent('ParkedCall', self.eventParkedMonitor)
		self.getConn().registerEvent('UnParkedCall', self.eventParkedMonitor)
		self.getConn().registerEvent('ParkedCallTimeOut', self.eventParkedMonitor)
		self.getConn().registerEvent('ParkedCallGiveUp', self.eventParkedMonitor)
		self.getConn().registerEvent('MeetmeJoin', self.eventConfMonitor)
		self.getConn().registerEvent('MeetmeLeave', self.eventConfMonitor)
		self.getConn().registerEvent('MeetmeEnd', self.eventConfMonitor)
		self.getConn().registerEvent('MessageWaiting', self.eventVoicemailMonitor)
		self.getConn().registerEvent('Join', self.eventQueueMonitor)
		self.getConn().registerEvent('Leave', self.eventQueueMonitor)
		self.getConn().registerEvent('QueueMemberAdded', self.eventQueueMonitor)
		self.getConn().registerEvent('QueueMemberPaused', self.eventQueueMonitor)
		self.getConn().registerEvent('QueueMemberRemoved', self.eventQueueMonitor)
		self.getConn().registerEvent('QueueMemberStatus', self.eventQueueMonitor) #Issued when agent logs-in/out and during call
		self.getConn().registerEvent('QueueParams', self.eventQueueMonitor)
		self.getConn().registerEvent('QueueMember', self.eventQueueMonitor) #Issued during call or reload
		self.getConn().registerEvent('Reload', self.eventReloadMonitor)
		#self.getConn().registerEvent( None, self.eventProcessor )
		self.syncTasks()
		QueueDB.ami = self #Pass reference to QueueDB so it can order queues sync, workaround to fix asterisk issue with not sending QueueMember
		

	def syncTasks( self ):
		"""
		Sync tasks here: allow 1 second in between so we do not hit unecessary peak loads
		"""
		tasks = [ self.syncSettings, self.syncSipPresence, self.syncConfs, self.syncQueues, 
				self.syncCallStatus, self.syncParked, self.syncAllMailboxes ]
		for idx in range(len(tasks)):
			reactor.callLater( idx, tasks[idx] ) # Call them normally, change 1 to idx in case we see peaks

	def syncSettings( self ):
		"""
		Used to keep cached all settings
		"""
		def onSipShowPeers( peer ):
			"""
			@param peer: a C{Dictionary} of peer settings, useful to extract peer context
			@example peer:
				{'busy-level': '0', 'status': 'OK (4 ms)', 'sip-sess-min': '90', 'sip-authinsecure': 'no', 
				'sip-canreinvite': 'N', 'maxcallbr': '384 kbps', 'call-limit': '2147483647', 'sip-rtp-engine': 'asterisk', 
				'amaflags': 'Unknown', 'sip-useragent': 'Jitsi1.0-build.3967Linux', 'mohsuggest': '', 'actionid': 'MbitDev-294431344-2', 
				'default-addr-ip': '(null)', 'callgroup': '', 'address-port': '5060', 'chanobjecttype': 'peer', 'objectname': '101', 
				'transfermode': 'open', 'sip-sess-timers': 'Accept', 'maxforwards': '0', 'sip-userphone': 'N', 'response': 'Success', 
				'sip-encryption': 'N', 'sip-textsupport': 'N', 'sip-directmedia': 'N', 'secretexist': 'Y', 
				'cid-callingpres': 'Presentation Allowed, Not Screened', 'codecs': '0xe (gsm|ulaw|alaw)', 'address-ip': '10.10.10.155', 
				'default-addr-port': '0', 'sip-sess-expires': '1800', 'dynamic': 'Y', 'parkinglot': '', 'qualifyfreq': '60000 ms', 
				'reg-contact': 'sip:101@10.10.10.155:5060;transport=udp;registering_acc=mbitdev', 'regexpire': '437 seconds', 
				'sip-sess-refresh': 'uas', 'sip-promiscredir': 'N', 'sip-forcerport': 'Y', 'lastmsgssent': '-1', 
				'sip-use-reason-header': 'N', 'channeltype': 'SIP', 'voicemailbox': '101@device', 'sip-t.38maxdtgrm': '-1', 
				'sip-t.38support': 'N', 'sip-videosupport': 'N', 'codecorder': 'ulaw,alaw,gsm', 'callerid': '"device" <101>', 
				'language': '', 'tohost': '', 'pickupgroup': '', 'acl': 'Y', 'sip-t.38ec': 'Unknown', 'context': 'from-internal', 
				'md5secretexist': 'N', 'default-username': '101', 'remotesecretexist': 'N', 'sip-dtmfmode': 'rfc2833'}
			"""
			for key in ("context", "voicemailbox"):
				if key == "voicemailbox":
					val = peer[key].split("@")[0] #only taking id and not context after @
					if not val: # Empty voicemailbox id, ignore
						continue
				else:
					val = peer[key]
				SettingsDB.setEntity( peer['objectname'], key, val ) #Save settings

		def onSipPeers( peers ):
			"""
			@param peers: a C{List} of C{Dictionary} representing peers
			@example peers:
				[{'ipport': '5060', 'status': 'OK (19 ms)', 'chanobjecttype': 'peer', 'objectname': '101', 
				'realtimedevice': 'no', 'dynamic': 'yes', 'textsupport': 'no', 'acl': 'yes', 
				'actionid': 'MbitDev-280414760-2', 'videosupport': 'no', 'channeltype': 'SIP', 
				'ipaddress': '10.10.10.155', 'event': 'PeerEntry', 'forcerport': 'yes'}]
			"""
			from time import sleep
			for peer in peers:
				self.getConn().sipShowPeers(peer['objectname']).addCallbacks( onSipShowPeers, log.err)

		def onShowKeyCIDName( res ):
			"""
			@example:
				['/AMPUSER/100/cidname                              : Dan100                   ', 
				'/AMPUSER/101/cidname                              : Dan101                   ', 
				'/AMPUSER/102/cidname                              : Dan102                   ', 
				'/AMPUSER/103/cidname                              : Dan103                   ', '4 results found.']
			"""
			for line in res:
				m = re.search('\/AMPUSER/(\w+)\/cidname\s+:\s(\w+\s*\w+)\s+', line)
				if not m:
					continue
				exten, name = m.group(1), m.group(2)
				SettingsDB.setEntity( exten, 'name', name )
		self.getConn().sipPeers().addCallbacks( onSipPeers, log.err )
		self.getConn().command("database showkey cidname").addCallbacks(onShowKeyCIDName, log.err) #Query extension names
		if self.settings['ami_sync']:
			reactor.callLater( self.settings['ami_sync'], self.syncSettings ) #Reschedule querying of the settings, to make sure we keep the sync


	def syncCallStatus( self, repeat=True ):
		def onStatus( channels ):
			"""
			Receives the channel status reply.
			@example channels:
				[{'bridgedchannel': 'SIP/101-00000003', 'connectedlinenum': '101', 'bridgeduniqueid': '1355502846.3', 
				'uniqueid': '1355502846.2', 'extension': 's', 'accountcode': '', 'channelstate': '6', 'calleridnum': '102', 
				'priority': '37', 'seconds': '284', 'connectedlinename': 'device', 'actionid': 'MbitDev-330060000-2', 
				'context': 'macro-dial-one', 'calleridname': 'Dan102', 'privilege': 'Call', 'event': 'Status', 
				'channel': 'SIP/102-00000002', 'channelstatedesc': 'Up'}, {'bridgedchannel': 'SIP/102-00000002', 
				'connectedlinenum': '102', 'account': '', 'bridgeduniqueid': '1355502846.2', 'calleridnum': '101', 
				'state': 'Up', 'connectedlinename': 'Dan102', 'actionid': 'MbitDev-330060000-2', 'uniqueid': '1355502846.3', 
				'calleridname': 'device', 'privilege': 'Call', 'event': 'Status', 'channel': 'SIP/101-00000003'}]
			"""
			CallsDB.syncCallStatus( channels )
		self.getConn().status().addCallbacks( onStatus, log.err )
		if self.settings['ami_sync'] and repeat:
			reactor.callLater( self.settings['ami_sync'], self.syncCallStatus )

	def syncSipPresence( self ):
		"""
		Extracts SIP presence by processing sip peers and querying their status
		"""
		def onSipPeers( peers ):
			"""
			@param peers: a C{List} of C{Dictionary} representing peers
			@example peers:
				[{'ipport': '5060', 'status': 'OK (19 ms)', 'chanobjecttype': 'peer', 'objectname': '101', 
				'realtimedevice': 'no', 'dynamic': 'yes', 'textsupport': 'no', 'acl': 'yes', 
				'actionid': 'MbitDev-280414760-2', 'videosupport': 'no', 'channeltype': 'SIP', 
				'ipaddress': '10.10.10.155', 'event': 'PeerEntry', 'forcerport': 'yes'}]
			"""
			for peer in peers:
				if peer['status'] == "UNKNOWN":
					if ExtensDB.getExtenStatus( peer["objectname"] ) !=  EXTEN_UNAVAILABLE: #Stored status differs from the one received
						ExtensDB.setExtenStatus( peer["objectname"], EXTEN_UNAVAILABLE )
				else:
					ctx = SettingsDB.getEntityProperty( peer['objectname'], "context", "default" )
					self.getConn().extensionState(peer['objectname'],ctx).addCallback(self.eventExtensionStatus).addErrback( log.err )
		self.getConn().sipPeers().addCallbacks( onSipPeers, log.err )
		if self.settings['ami_sync']:
			reactor.callLater( self.settings['ami_sync'], self.syncSipPresence ) #Reschedule querying of the presence, to make sure we keep the sync


	def syncParked( self ):
		"""
		@example parked:
			[{'connectedlinenum': '101', 'from': 'SIP/101-00000002', 'exten': '701', 'calleridnum': '102', 
			'parkinglot': 'default', 'connectedlinename': 'Dan101', 'actionid': 'MbitDev-174490184-4', 
			'timeout': '54', 'calleridname': 'device', 'event': 'ParkedCall', 'channel': 'SIP/102-00000003'}]
		"""
		self.getConn().parkedCalls().addCallbacks( ParkDB.syncParkedCalls, log.err )
		if self.settings['ami_sync']:
			reactor.callLater( self.settings['ami_sync'], self.syncParked ) #Reschedule querying of parked calls to keep in sync


	def eventParkedMonitor( self, conn, event ):
		"""
		@example event:
			{'connectedlinenum': '101', 'from': 'SIP/101-00000008', 'uniqueid': '1356278605.11', 'exten': '701', 
			'calleridnum': '102', 'parkinglot': 'default', 'connectedlinename': 'Dan101', 'timeout': '60', 
			'calleridname': 'device', 'privilege': 'call,all', 'event': 'ParkedCall', 'channel': 'SIP/102-00000009'}
			{'connectedlinenum': '101', 'from': 'SIP/101-0000001b', 'exten': '701', 'calleridnum': '102', 'parkinglot': 'default', 
			'connectedlinename': 'Dan101', 'uniqueid': '1356283146.34', 'calleridname': 'device', 'privilege': 'call,all', 
			'event': 'UnParkedCall', 'channel': 'SIP/102-0000001a'}
			{'connectedlinenum': '101', 'exten': '701', 'calleridnum': '102', 'parkinglot': 'default', 'connectedlinename': 'Dan101', 
			'uniqueid': '1356282965.30', 'calleridname': 'device', 'privilege': 'call,all', 'event': 'ParkedCallTimeOut', 
			'channel': 'SIP/102-00000017'}
		"""
		if event['event'] == 'ParkedCall':
			ParkDB.parked( event )
		elif event['event'] in ['UnParkedCall', 'ParkedCallTimeOut', 'ParkedCallGiveUp']:
			ParkDB.unparked(event['channel'])
		else:
			log.msg("WARNING: %s - Unsupported event type %s"%self.__class__.__name__, event['event'])


	def syncConfs( self ):
		def onConfigs( allConfigs ):
			"""
			Called with all configs in the system. Will extract useful info from app_meetme related here
			"""
			for roomNr in libmbit.confRoomsFromConfigs( allConfigs ):
				ConfDB.addConfRoom(roomNr) # Wll be added if not already in
			for room in ConfDB.getConfRooms():
				self.getConn().command("meetme list %s"%room).addCallback(onMeetMeList, room)

		def onMeetMeList( cmdResult, room ):
			"""
			Called with the answer from command: "meetme list confRoomNr"
			"""
			membersData = libmbit.confRoomParticipants( cmdResult ) #@example membersData: [[1, '101', 'SIP/101-00000003', 1361790916.0]]
			ConfDB.syncMembersFromListCmd( room, membersData )

		self.getConn().command("config list").addCallback( onConfigs )

		if self.settings['ami_sync']:
			reactor.callLater( self.settings['ami_sync'], self.syncConfs )

	def eventReloadMonitor( self, conn, event ):
		"""
		Called when a reload was requested, most of the cases when reconfiguration happens
		"""
		self.syncTasks()


	def eventConfMonitor( self, conn, event ):
		"""
		Receives events related to conference
		"""
		# {'connectedlinenum': '<unknown>', 'usernum': '1', 'meetme': '5000', 'calleridnum': '101', 'connectedlinename': '<unknown>', 'uniqueid': '1356545386.102', 'calleridname': 'Dan101', 'privilege': 'call,all', 'event': 'MeetmeJoin', 'channel': 'SIP/101-00000041'}
		# {'connectedlinenum': '<unknown>', 'usernum': '1', 'meetme': '5000', 'calleridnum': '101', 'duration': '215', 'connectedlinename': '<unknown>', 'uniqueid': '1356545386.102', 'calleridname': 'Dan101', 'privilege': 'call,all', 'event': 'MeetmeLeave', 'channel': 'SIP/101-00000041'}
		# {'privilege': 'call,all', 'event': 'MeetmeEnd', 'meetme': '5000'}
		if event['event'] == 'MeetmeJoin':
			ConfDB.joinConf( event['meetme'], event['channel'], event['calleridnum'], event['usernum'] )
		elif event['event'] == 'MeetmeLeave':
			ConfDB.leaveConf( event['meetme'], event['channel'] )
		elif event['event'] == 'MeetmeEnd':
			ConfDB.endConf( event['meetme'] )

	def eventProcessor( self, conn, event ):
		log.msg( "New event received:%s"%event )

	def eventCallMonitor( self, conn, event ):
		#{'accountcode': '', 'uniqueid': '1363047360.8', 'channelstate': '0', 'exten': '', 'calleridnum': '101', 'context': 'from-internal', 'calleridname': 'device', 'privilege': 'call,all', 'event': 'Newchannel', 'channel': 'SIP/101-00000008', 'channelstatedesc': 'Down'}
		print "eventCallMonitor", event
		if event['event'] == 'Newstate':
			# {'connectedlinenum': '101', 'channelstate': '6', 'calleridnum': '102', 'connectedlinename': 'device', 
			# 'uniqueid': '1355503396.4', 'calleridname': 'Dan102', 'privilege': 'call,all', 'event': 'Newstate', 
			# 'channel': 'SIP/102-00000004', 'channelstatedesc': 'Up'}
			CallsDB.callUpdate(event['channel'], event['calleridnum'], event['connectedlinenum'], event['channelstate'])
		elif event['event'] == 'Newchannel': #On two legs call exten attribute is empty, sort of workaround
			# {'accountcode': '', 'uniqueid': '1355503396.4', 'channelstate': '0', 'exten': '101', 'calleridnum': '102', 
			# 'context': 'from-internal', 'calleridname': 'device', 'privilege': 'call,all', 'event': 'Newchannel', 
			# 'channel': 'SIP/102-00000004', 'channelstatedesc': 'Down'}
			if event['exten'] != "":
				calling = event['calleridnum']
				called = event['exten']
			else: #Extension is empty, second leg call, swap calling/called
				calling = event['exten']
				called = event['calleridnum']
			CallsDB.callStart(event['channel'], calling, called, event['context'])
		elif event['event'] == 'Dial' and event['subevent']=='Begin':
			# {'connectedlinenum': '<unknown>', 'destination': 'SIP/102-0000000d', 'dialstring': '102', 
			# 'calleridnum': '101', 'destuniqueid': '1355764065.13', 'subevent': 'Begin', 'connectedlinename': '<unknown>', 	
			# 'uniqueid': '1355764065.12', 'calleridname': 'Dan101', 'privilege': 'call,all', 'event': 'Dial', 'channel': 'SIP/101-0000000c'}
			CallsDB.callDialed( event['channel'], event['calleridnum'], event['dialstring'], event['destination'] )
		#{'dialstatus': 'ANSWER', 'subevent': 'End', 'uniqueid': '1363063394.3', 'privilege': 'call,all', 'event': 'Dial', 'channel': 'SIP/101-00000003'}
		#elif event['event'] == 'Dial' and event['subevent'] == 'End':
		#	CallsDB.callStop(event['channel'])
		#ToDo: Patch here for unparking calls to show up
		elif event['event'] == 'Hangup':
			#{'connectedlinenum': '102', 'cause-txt': 'Normal Clearing', 'calleridnum': '101', 'connectedlinename': 'Dan102', 
			#'uniqueid': '1355503397.5', 'calleridname': 'device', 'privilege': 'call,all', 'cause': '16', 
			# 'event': 'Hangup', 'channel': 'SIP/101-00000005'}
			CallsDB.callStop(event['channel'])

	def eventExtensionStatus( self, *args ):
		"""
		@type event: dictionary of event status
		@example event:
			# Automated
			{'status': '0', 'hint': 'SIP/101', 'exten': '101', 'context': 'ext-local', 'privilege': 'call,all', 'event': 'ExtensionStatus'}
			#From query:
			{'status': '1', 'hint': 'SIP/101', 'exten': '101', 'actionid': 'MbitDev-259877416-2', 
				'context': 'from-internal', 'message': 'Extension Status', 'response': 'Success'}
		"""
		# Detect whether event is generated automatically or calledback from extensionState command
		if len(args)==2: # From event
			event = args[1] #First arg is connection
			#voicemail = args[2]
		elif len(args)==1: # From query
			event = args[0]
		event["status"] # Contains the status info we are interested in
		if ExtensDB.getExtenStatus( event["exten"] ) !=  event["status"]: #Stored status differs from the one received, update db
			ExtensDB.setExtenStatus( event["exten"], event["status"] )


	def eventVoicemailMonitor( self, conn, event ):
		"""
		@example:
			{'old': '2', 'mailbox': '102@default', 'waiting': '1', 'privilege': 'call,all', 'new': '1', 'event': 'MessageWaiting'}
			{'privilege': 'call,all', 'waiting': '0', 'event': 'MessageWaiting', 'mailbox': '102@default'}
		"""
		self.syncMailbox( event['mailbox'] )
		

	def syncMailbox( self, vmid ):
		def onMailboxCount( event ):
			"""
			@example: {'mailbox': '102@default', 'newmessages': '0', 'message': 'Mailbox Message Count', 'urgmessages': '0', 
				'oldmessages': '3', 'response': 'Success', 'actionid': 'MbitDev-57945192-5'}
			"""
			MailboxDB.setMailbox(event['mailbox'].split("@")[0], event['newmessages'], event['oldmessages'], event['urgmessages'])
		self.getConn().mailboxCount(vmid).addCallbacks( onMailboxCount, log.err )


	def syncAllMailboxes( self ):
		for vmid in SettingsDB.getPropertyValues('voicemailbox'):
			self.syncMailbox( vmid )


	def eventQueueMonitor( self, conn, event ):
		"""
		Called when an event related to queues fires up
		@example join:
			{'count': '1', 'connectedlinenum': 'unknown', 'calleridnum': '102', 'queue': '200', 'connectedlinename': 'unknown', 
			'uniqueid': '1356823539.12', 'calleridname': '123Dan102', 'privilege': 'call,all', 'position': '1', 'event': 'Join', 
			'channel': 'SIP/102-0000000c'}
		@example leave:
			{'count': '0', 'queue': '200', 'uniqueid': '1356823539.12', 'privilege': 'call,all', 'position': '1', 
			'event': 'Leave', 'channel': 'SIP/102-0000000c'}
		@example abandon:
			{'queue': '200', 'originalposition': '1', 'uniqueid': '1356823539.12', 'privilege': 'agent,all', 'position': '1', 
			'holdtime': '47', 'event': 'QueueCallerAbandon'}
		@example member:
			{'status': '1', 'penalty': '0', 'paused': '0', 'name': 'Dan102', 'queue': '200', 'membership': 'static', 
			'location': 'Local/102@from-queue/n', 'lastcall': '1356828718', 'event': 'QueueMember', 'callstaken': '2', 
			'actionid': 'MbitDev-350983808-2'}
		@example memberAdded:
			{'status': '1', 'penalty': '0', 'paused': '1', 'membername': 'sip/103', 'queue': '200', 'lastcall': '0', 
			'membership': 'dynamic', 'location': 'sip/103', 'privilege': 'agent,all', 'event': 'QueueMemberAdded', 'callstaken': '0'}
		@example memberPaused:
			{'paused': '1', 'membername': 'sip/103', 'queue': '200', 'location': 'sip/103', 'privilege': 'agent,all', 
			'event': 'QueueMemberPaused'}
		@example memberRemoved:
			{'privilege': 'agent,all', 'queue': '200', 'membername': 'sip/103', 'event': 'QueueMemberRemoved', 'location': 'sip/103'}
		@example memberStatus:
			{'status': '1', 'penalty': '0', 'paused': '0', 'membername': 'Dan101', 'queue': '200', 'lastcall': '0', 'membership': 'static', 
				'location': 'Local/101@from-queue/n', 'privilege': 'agent,all', 'event': 'QueueMemberStatus', 'callstaken': '0'}
		"""
		if event['event'] == 'Join':
			QueueDB.join( event['queue'], event['position'], event['calleridnum'], event['channel'], 0 )
			#self.syncQueues( repeat=False, queue=event['queue'] ) #Need to refresh here also since not all extensions have presence
		elif event['event'] == 'Leave':
			QueueDB.leave( event['queue'], event['channel'] )
			#self.syncQueues( repeat=False, queue=event['queue'] )
		elif event['event'] in ('QueueMember','QueueMemberAdded') :
			QueueDB.setAgent( event['queue'], libmbit.entityFromLocation(event['location']), event['membership'], event['status'],
				event['location'], event['callstaken'], event['lastcall'], event['paused'], event['penalty'] )
		elif event['event'] == 'QueueMemberPaused':
			QueueDB.pauseAgent( event['queue'], libmbit.entityFromLocation(event['location']), event['paused'] )
		elif event['event'] == 'QueueMemberRemoved':
			QueueDB.rmAgent( event['queue'], libmbit.entityFromLocation(event['membername']) )
		elif event['event'] == 'QueueMemberStatus':
			QueueDB.setAgent( event['queue'], libmbit.entityFromLocation(event['location']), event['membership'], event['status'],
				event['location'], event['callstaken'], event['lastcall'], event['paused'], event['penalty'] )
			
	def syncQueues( self, repeat=True, queue=None, member=None ):
		"""
		@ToDo: call it whenever there is a change in status of an agent
		@param repeat: used whenever manual sync is called
		"""
		def onQueueStatus( queueStatus ):
			"""
			@example queueStatus:
				[{'servicelevelperf': '0.0', 'talktime': '0', 'calls': '0', 'max': '0', 'completed': '0', 'servicelevel': '0', 
				'weight': '0', 'strategy': 'ringall', 'queue': 'default', 'actionid': 'MbitDev-409077376-2', 'holdtime': '0', 
				'event': 'QueueParams', 'abandoned': '0'}, 
				{'servicelevelperf': '100.0', 'talktime': '16', 'calls': '0', 'max': '0', 'completed': '2', 'servicelevel': '60', 
				'weight': '0', 'strategy': 'ringall', 'queue': '200', 'actionid': 'MbitDev-409077376-2', 'holdtime': '2', 
				'event': 'QueueParams', 'abandoned': '23'}, 
				{'status': '1', 'penalty': '0', 'paused': '0', 'name': 'Dan102', 'queue': '200', 'membership': 'static', 
				'location': 'Local/102@from-queue/n', 'lastcall': '0', 'event': 'QueueMember', 'callstaken': '0', 
				'actionid': 'MbitDev-409077376-2'}]
			"""
			for qDetail in queueStatus:
				if qDetail['event'] == 'QueueEntry':
					QueueDB.join( qDetail['queue'], qDetail['position'], qDetail['calleridnum'], 
						qDetail['channel'],qDetail['wait'] )
				elif qDetail['event'] == 'QueueParams':
					stats = {}
					for key in ('servicelevelperf', 'talktime', 'calls', 'max', 'completed', 'servicelevel', 
						'weight', 'strategy',  'holdtime', 'abandoned'):
						stats[key] = qDetail[key]
					QueueDB.setStats( qDetail['queue'], stats )
				elif qDetail['event'] == 'QueueMember':
					entity = libmbit.entityFromLocation(qDetail['location'])
					if not entity:
						log.msg("WARNING: %s - Cannot extract entity from location %s"%(self.__class__.__name__, qDetail['location']))
						continue
					QueueDB.setAgent( qDetail['queue'], entity, qDetail['membership'], qDetail['status'], 
						qDetail['location'], qDetail['callstaken'], qDetail['lastcall'], qDetail['paused'], qDetail['penalty'] )
		self.getConn().queueStatusFiltered(queue, member).addCallbacks( onQueueStatus, log.err )
		if repeat and self.settings['ami_sync']:
			reactor.callLater( self.settings['ami_sync'], self.syncQueues ) #Reschedule querying of the settings, to make sure we keep the sync


