from time import time
from twisted.python import log
from twisted.internet import reactor

from mbitrans import libmbit
from mbitrans.settingsdb import SettingsDB

CALL_STATES = {
	"Down": "0",
	"Rsrvd": "1",
	"OffHook": "2",
	"Dialing": "3",
	"Ring": "4",
	"Ringing": "5",
	"Up": "6",
	"Busy": "7",
	"Dialing Offhook": "8",
	"Pre-ring": "9",
	"Unknown": "10",
	}


class CallsDb( object ):
	"""I am responsible to keep a cache of active calls on the platform
	"""
	config = {} # Will be populated here by tap module
        #{'SIP/211-00000054': ['211', '212', '6', 0, 1363083752.941402, 'SIP/212-00000055']}
	activeCalls = {} # Indexed calls per Ids, in the form of {'aChanId':[from, to, chanStatus, setupTime, connectTime, context, bChanId, callLeg]}
	extensionCalls = {}  # Map extension to list of tuple in the form id, direction {'101:[('someid','IN'),('otherid','OUT')]}
	statusNotifiers  = set()

	def registerStatusNotifier( self, notifier ):
		"""External functions can use this method to register callback in case of calls db changes"""
		self.statusNotifiers.add( notifier )

	def unregisterStatusNotifier( self, notifier ):
		"""Unregister from receiving notifies on database changes"""
		self.statusNotifiers.discard( notifierMethod )

	def triggerStatusNotifies( self, extension ):
		"""Registered functions will be called to inform about presence changes on specific extension"""
		for notifier in self.statusNotifiers:
			reactor.callLater(self.config['gateway_delay'], notifier, extension)

	def srchBridgedChanId( self, srchChan ):
		"""
		Searches through the active calls matching the bridged channel
		"""
		for chanId in self.activeCalls.keys():
			if self.activeCalls[chanId][6] == srchChan:
				return chanId
		return ""

	def attachCallToExtensions( self, callLeg, chanId, calling, called ):
		"""
		Attach call to extension so it can be retrieved properly
		"""
		if callLeg == "A" and calling:
			if calling in self.extensionCalls:
				self.extensionCalls[calling].add( (chanId, 'OUT') )
			else:
				self.extensionCalls[calling] = set([(chanId, 'OUT')])
			self.triggerStatusNotifies( calling )
		elif callLeg == "B" and called:
			if called in self.extensionCalls:
				self.extensionCalls[called].add( (chanId, 'IN') )
			else:
				self.extensionCalls[called] = set([(chanId, 'IN')])
			self.triggerStatusNotifies( called )

	def detachCallFromExtensions( self, chanId ):
		"""
		Detach call from extension, useful on updates
		"""
		for exten in self.extensionCalls.keys():
			for callInfo in set(self.extensionCalls[exten]):
				if callInfo[0] == chanId:
					self.extensionCalls[exten].discard(callInfo)
					self.triggerStatusNotifies( exten )
                       		if len(self.extensionCalls[exten]) == 0:
                               		del self.extensionCalls[exten]

	def callStart( self, chanId, calling, called, context, bChanId='', callLeg='' ):
		"""
		@param chanId: originator chanId
		@param calling: calling extension
		@param called: called extension
		@param context: call context
		"""
		if chanId.find("from-queue-") != -1: #Ignore calls originated wihin queue since they produce triple records
			return
		if not callLeg:
			if calling:
				callLeg = "A"
			else:
				callLeg = "B"
		self.activeCalls[chanId] = [calling, called, "-1", time(), 0, context, bChanId, callLeg]
		self.attachCallToExtensions( callLeg, chanId, calling, called )
		print "callStart", self.activeCalls
		print "callStart", self.extensionCalls

	def callDialed( self, chanId, calling, called, bChanId ):
		"""
		Useful when originate is done over cmd, adds bChanId
		@param chanId: originator call identity
		{'connectedlinenum': '<unknown>', 'destination': 'SIP/211-0000002f', 'dialstring': '211', 'calleridnum': '0280171539', 'destuniqueid': '1364574132.56', 'subevent': 'Begin', 'connectedlinename': '<unknown>', 'uniqueid': '1364574132.55', 'calleridname': '0280171539', 'privilege': 'call,all', 'event': 'Dial', 'channel': 'SIP/8467768-0000002e'}
		SIP/101-00000000 101 102 SIP/102-00000001
		"""
		print "callDialed", chanId, calling, called, bChanId
		if chanId and chanId not in self.activeCalls:
			ctx = SettingsDB.getEntityProperty( calling, "context", "default" )
			self.callStart( chanId, calling, called, ctx, bChanId, "A" )
		elif bChanId and bChanId not in self.activeCalls:
			ctx = SettingsDB.getEntityProperty( calling, "context", "default" )
			self.callStart( bChanId, calling, called, ctx, "", "B" )
		for chan in (chanId, bChanId):
			if chan != bChanId and self.activeCalls[chan][6]!=bChanId: #chanId not having bChanId recorded in leg cdr, record it
				self.activeCalls[chan][6] = bChanId
			callLeg = self.activeCalls[chan][7]
			if callLeg == "B": #emulate behavior of normal event with swapped subjects
				self.callUpdate( chan, called, calling, "" )
			else:
				self.callUpdate( chan, calling, called, "" )
		if self.activeCalls[chanId][6] != bChanId: #correct bChanId of the calling leg
			self.activeCalls[chanId][6] = bChanId
		if self.activeCalls[bChanId][6] != chanId: #correct bChanId of the called leg
			self.activeCalls[bChanId][6] = chanId


	def callUpdate( self, chanId, calling, called, chanStatus ):
		"""
		@param chanId: originator chanId
		@param chanStatus: chanStatus
		[calling, called, "-1", time(), 0, context, bChanId, callLeg]
		"""
		print "callUpdate", chanId, calling, called, chanStatus
		if chanId not in self.activeCalls:
			return
		oCalling, oCalled, oStat, callLeg = self.activeCalls[chanId][0], self.activeCalls[chanId][1], self.activeCalls[chanId][2], self.activeCalls[chanId][7]
		if callLeg == "B": #Swap calling with called due to way asterisks sees them
			calling, called = called, calling
		if (oCalling, oCalled, oStat) == (calling, called, chanStatus):
			return # No change
		if calling and oCalling != calling:
			self.activeCalls[chanId][0] = calling
		if called and oCalled != called:
			self.activeCalls[chanId][1] = called
		if chanStatus and oStat != chanStatus:
			self.activeCalls[chanId][2] = chanStatus
		if callLeg == "B" and calling in self.extensionCalls.keys() and ((chanId, 'IN') in self.extensionCalls[calling]):
			# Special case by by originate when both ends of an originate are with "B" call leg, however callDialed fixes it
			self.activeCalls[chanId][7] = 'A'
			self.detachCallFromExtensions( chanId )
			self.attachCallToExtensions( self.activeCalls[chanId][7], chanId, calling, called )
		if chanStatus == '6': 
			self.activeCalls[chanId][4] = time() #Channel connected, record connect time
			if self.activeCalls[chanId][7] == "A":
				srchBChanId = self.srchBridgedChanId( chanId )
				if srchBChanId and self.activeCalls[chanId][6] != srchBChanId:
					self.activeCalls[chanId][6] = srchBChanId #For a leg b chanid is only confirmed at connect time,eg. group calling
		# Good to go, notify changes
		if self.activeCalls[chanId][7] == "A":
			self.triggerStatusNotifies( calling )
		elif self.activeCalls[chanId][7] == "B":
			self.triggerStatusNotifies( called )
		print "callUpdate", self.activeCalls
		print "callUpdate", self.extensionCalls


	def callStop( self, chanId ):
		"""
		same params as for L{self.callStart}
		"""
		if not chanId in self.activeCalls:
			return # Nothing to do since call is not in cache
		self.detachCallFromExtensions( chanId )
		del self.activeCalls[chanId]
		print "callStop", self.activeCalls
		print self.extensionCalls

	def callSync( self, chanId, calling, called, context, chanStatus, dur, bChanId, callLeg):
		if chanId not in self.activeCalls:
			self.activeCalls[chanId] = [calling, called, chanStatus, time(), time()-int(dur), context, bChanId, callLeg]
			self.attachCallToExtensions( callLeg, chanId, calling, called )
		elif chanStatus != self.activeCalls[chanId][2]:
			self.callUpdate( chanId, calling, called, chanStatus )
		
	def syncCallStatus( self, amiChansInfo ):
		"""
		Called by AMI with active chanIds retreieved from status command
		@param amiChansInfo: list of dictionaries, each representing a channel
		@example for one call in+out:
		[{'bridgedchannel': 'SIP/103-00000006', 'connectedlinenum': '103', 'account': '', 'bridgeduniqueid': '1363004999.6', 'calleridnum': '102', 
			'state': 'Up', 'connectedlinename': 'Dan103', 'actionid': 'MbitDev-372547736-6', 'uniqueid': '1363004999.7',  
			'calleridname': 'device', 'privilege': 'Call', 'event': 'Status', 'channel': 'SIP/102-00000007'}, 
		{'bridgedchannel': 'SIP/102-00000007', 'connectedlinenum': '102', 'bridgeduniqueid': '1363004999.7', 'uniqueid': '1363004999.6', 
			'extension': 's', 'accountcode': '', 'channelstate': '6', 'calleridnum': '103', 'priority': '37', 'seconds': '15', 
			'connectedlinename': 'device', 'actionid': 'MbitDev-372547736-6', 'context': 'macro-dial-one', 'calleridname': 'Dan103', 
			'privilege': 'Call', 'event': 'Status', 'channel': 'SIP/103-00000006', 'channelstatedesc': 'Up'}]
		"""
		print "syncing calls", amiChansInfo
		print "on syncing have calls in db", self.activeCalls
		print "on syncing have extension calls", self.extensionCalls
		# Cleanup first
		amiChanIds = [ chanInfo["channel"] for chanInfo in amiChansInfo ]
		for chanId in self.activeCalls.keys():
			if chanId not in amiChanIds:
				self.callStop( chanId )
		# Add new calls now
		for chanInfo in amiChansInfo:
			if 'extension' in chanInfo: # Second leg does not have extension and calling is swapped
				calling = chanInfo['calleridnum']
				called = chanInfo['connectedlinenum']
				chanState = chanInfo['channelstate']
				dur = chanInfo['seconds']
				callLeg = "A"
			else:
				calling = chanInfo['connectedlinenum']
				called = chanInfo['calleridnum']
				chanState = "-1" # We should take it out of state which is not code, no need to support it now
				if chanInfo["state"] in CALL_STATES:
					chanState = CALL_STATES[chanInfo["state"]]
				dur = 0
				callLeg = "B"
			if 'context' in chanInfo:
				ctx = chanInfo['context']
			else:
				ctx = SettingsDB.getEntityProperty( calling, 'context', 'default' )
			bChanId = ""
			if 'bridgedchannel' in chanInfo: #Sometimes (eg by queues) this info is not in the sync data
				bChanId = chanInfo['bridgedchannel']
			self.callSync( chanInfo['channel'], calling, called, ctx, chanState, dur, bChanId, callLeg )
			

	def getExtensionCalls( self, exten ):
		"""
		Returns calls active for an extension
		@returns: call info in the form:[ [direction, calling, called, chanState, setupTime, connectTime], ]
		"""
		extenCalls = []
		if exten in self.extensionCalls.keys():
			for call in self.extensionCalls[exten]:
				extenCalls.append([call[0], call[1]] + self.activeCalls[call[0]])
		return extenCalls


CallsDB = CallsDb()

