"""
Commands handler
"""
from wokkel.subprotocols import XMPPHandler, IQHandlerMixin
from wokkel.data_form import Form, Field
from wokkel.disco import DiscoIdentity
from twisted.internet import defer
from twisted.words.protocols.jabber import error
from twisted.python import log
from twisted.internet import reactor
from starpy.error import AMICommandFailure
from mbitrans import adhoc
from mbitrans.callsdb import CallsDB
from mbitrans.parkdb import ParkDB
from mbitrans.extensdb import ExtensDB
from mbitrans.permsdb import PermsDB
from mbitrans.mailboxdb import MailboxDB
from mbitrans.confdb import ConfDB
from mbitrans.queuedb import QueueDB
from mbitrans.settingsdb import SettingsDB

CMD_WITH_NS = "/command[@xmlns='%s']" % adhoc.NS_ADHOC
XPATH_CMD_GET = "/iq[@type='get']" + CMD_WITH_NS
XPATH_CMD_SET = "/iq[@type='set']" + CMD_WITH_NS
XPATH_CMD_GET_OR_SET = "/iq[@type='get' or @type='set']" + CMD_WITH_NS


class CmdHandler(XMPPHandler, IQHandlerMixin):
	"""
	Takes care about all commands received from xmpp side
	Dispatches commands via "run_%s"%verb methods
	"""

	
	iqHandlers = {XPATH_CMD_SET: 'cmdDispatcher'}

	discoIdentity = DiscoIdentity('commands',
				'service',
				'Adhoc Commands Service')

	# ToDo: Implement me to comply with IDisco
	def getDiscoInfo(self, requestor, target, nodeIdentifier=''):
		pass
	# ToDo: Implement me to comply with IDisco
	def getDiscoItems(self, requestor, target, nodeIdentifier=''):
		pass

	def connectionMade(self):
		self.xmlstream.addObserver( XPATH_CMD_SET, self.handleRequest )

	def cmdDispatcher( self, iq ):
		"""
		Called on each command
		"""
		cmd = adhoc.Command.fromElement( iq.firstChildElement() )
		try:
			runcmd = getattr( self, "run_%s"%cmd.command )
		except AttributeError:
			return defer.fail( error.StanzaError( 'not-acceptable' ) )
		from time import sleep
		sleep(0.005)
		return runcmd( cmd )

	def run_version( self, cmd ):
		"""
		Sample command to execute
		@type cmd: adhoc.Command
		"""
		print "Version called"

	def run_hangup( self, cmd ):
		"""
		Hangups a running call
		"""
		if 'chanid' not in cmd.form.fields or len(cmd.form.fields['chanid'].values)==0 or len(cmd.form.fields['chanid'].values[0])==0:
			return defer.fail( error.StanzaError( 'not-acceptable' ) )
		return self.parent.runAMICommand( 'hangup', channel=cmd.form.fields['chanid'].values[0] ).addCallback( lambda msg: None )

	def run_originate( self, cmd ):
		"""
		Originates a call, requires from and to fields
		"""
		for field in ('from', 'to', ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		return self.parent.runAMICommand( 'originate', channel="sip/%s"%cmd.form.fields['from'].values[0], 
				priority='1', exten=cmd.form.fields['to'].values[0]).addCallback( lambda msg: None ) #absorbe response

	def run_transfer( self, cmd ):
		"""
		Transfer an existing call
		"""
		for field in ('chanid', 'to', 'user' ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		chanid = cmd.form.fields['chanid'].values[0]
                if  chanid not in CallsDB.activeCalls:
                        return defer.fail( error.StanzaError( 'not-acceptable' ) )
		return self.parent.runAMICommand('redirect', channel=chanid,
				context= SettingsDB.getEntityProperty(cmd.form.fields['user'].values[0], "context", "default"),
				exten=cmd.form.fields['to'].values[0], priority='1').addCallback( lambda msg: None )


	def run_park( self, cmd ):
		"""
		Transfer an existing call
		"""
		for field in ('chanid', ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		chanid = cmd.form.fields['chanid'].values[0]
		if  chanid not in CallsDB.activeCalls or not CallsDB.activeCalls[chanid][6]:
			return defer.fail( error.StanzaError( 'not-acceptable' ) )
		return self.parent.runAMICommand( 'park', channel=chanid, channel2=CallsDB.activeCalls[chanid][6], timeout=60000 ).addCallback( lambda msg: None )
	
	def run_unpark( self, cmd ):
		"""
		Retrieves a parked call.
		Work around to sync calls since Asterisk does not report call bridged with unpark properly
		"""
		for field in ('chanid', 'user', ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		chanid = cmd.form.fields['chanid'].values[0]
		if chanid not in ParkDB.parkedCalls:
			return defer.fail( error.StanzaError( 'not-acceptable' ) )
		self.parent.runAMICommand( 'originate', channel="sip/%s"%cmd.form.fields['user'].values[0], 
				priority='1', exten=ParkDB.parkedCalls[chanid][1])
		reactor.callLater(2, self.parent.runAMIFctryCommand, 'syncCallStatus', **{"repeat":False})
		reactor.callLater(10, self.parent.runAMIFctryCommand, 'syncCallStatus', **{"repeat":False})



	def run_record( self, cmd ):
		"""
		Record an existing channel
		"""
		for field in ('action', 'chanid', 'user'): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		chanid = cmd.form.fields['chanid'].values[0]
		if  chanid not in CallsDB.activeCalls:
			return defer.fail( error.StanzaError( 'not-acceptable' ) )
		if cmd.form.fields['action'].values[0] == 'start':
			return self.parent.runAMICommand( 'monitor', channel=chanid, file='', format='', mix='1' ).addCallback( lambda msg: None )
		elif cmd.form.fields['action'].values[0] == 'pause':
			return self.parent.runAMICommand( 'pauseMonitor', channel=chanid ).addCallback( lambda msg: None )
		elif cmd.form.fields['action'].values[0] == 'unpause':
			return self.parent.runAMICommand( 'unpauseMonitor', channel=chanid ).addCallback( lambda msg: None )
		elif cmd.form.fields['action'].values[0] == 'stop':
			return self.parent.runAMICommand( 'stopMonitor', channel=chanid ).addCallback( lambda msg: None )
		else: #Unrecognized action
			return defer.fail( error.StanzaError( 'not-acceptable' ) )


	def run_spy( self, cmd ):
		"""
		Spy on a channel
		"""
		for field in ('chanid', 'user'): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		chanid = cmd.form.fields['chanid'].values[0]
		if  chanid not in CallsDB.activeCalls:
			return defer.fail( error.StanzaError( 'not-acceptable' ) )
		data = chanid
		if 'whisper' in cmd.form.fields:
			data += ',w'
		return self.parent.runAMICommand('originate', channel="sip/%s"%cmd.form.fields['user'].values[0], 
				application='ChanSpy', data=data).addCallback( lambda msg: None )

	def run_leave_voicemail( self, cmd ):
		"""
		Leaves a voicemail to some user
		"""
		for field in ('to', 'user'): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				return defer.fail( error.StanzaError( 'not-acceptable' ) )
		return self.parent.runAMICommand( 'originate', channel="sip/%s"%cmd.form.fields['user'].values[0], 
				priority='1', exten='*%s'%cmd.form.fields['to'].values[0]).addCallback( lambda msg: None )

	def run_conference( self, cmd ):
		"""
		Mutes a participant in a conference
		"""
		for field in ('confnr', 'action', 'usernr', ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				#return defer.fail( error.StanzaError( 'not-acceptable' ) )
				raise error.StanzaError( 'not-acceptable' )
		if cmd.form.fields['action'].values[0] == 'mute':
			return self.parent.runAMICommand( 'meetmeMute' , meetme=cmd.form.fields['confnr'].values[0], 
				usernum=cmd.form.fields['usernr'].values[0]).addCallback( lambda msg: None )
		elif cmd.form.fields['action'].values[0] == 'unmute':
			return self.parent.runAMICommand( 'meetmeUnmute' , meetme=cmd.form.fields['confnr'].values[0], 
				usernum=cmd.form.fields['usernr'].values[0] ).addCallback( lambda msg: None )

	def run_queue_reset( self, cmd ):
		"""
		Mutes a participant in a conference
		"""
		def onQueueReset(result):
			"""
			Since Asterisk does not refresh automatically stats, we need to do it here
			"""
			ami = self.parent.parent.getServiceNamed("AMI")._connection.factory
			ami.syncQueues( repeat=False )
			
		for field in ('queue', ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				#return defer.fail( error.StanzaError( 'not-acceptable' ) )
				raise error.StanzaError( 'not-acceptable' )
		return self.parent.runAMICommand( 'queueReset' , queue=cmd.form.fields['queue'].values[0]).addCallbacks( onQueueReset, log.err )

	def run_agent_logoff( self, cmd ):
		for field in ('agent', ): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				#return defer.fail( error.StanzaError( 'not-acceptable' ) )
				raise error.StanzaError( 'not-acceptable' )
		return self.parent.runAMICommand( 'agentLogoff' , agent=cmd.form.fields['agent'].values[0], soft=True ).addCallback( lambda msg: None )


	def run_queue_member( self, cmd ):
		"""
		Add/Pause/Remove a member interface to a queue
		"""
		for field in ('queue', 'member', 'action'): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				#return defer.fail( error.StanzaError( 'not-acceptable' ) )
				raise error.StanzaError( 'not-acceptable' )
		action=cmd.form.fields['action'].values[0]
		if action == 'add':
			penalty = 0
			if 'penalty' in cmd.form.fields:
				penalty = int(cmd.form.fields['penalty'].values[0])
			return self.parent.runAMICommand( 'queueAdd' , queue=cmd.form.fields['queue'].values[0],
				interface='sip/%s'%cmd.form.fields['member'].values[0], penalty=penalty ).addCallback( lambda msg: None )
		elif action == 'pause':
			return self.parent.runAMICommand( 'queuePause' , queue=cmd.form.fields['queue'].values[0], 
				interface='sip/%s'%cmd.form.fields['member'].values[0] ).addCallback( lambda msg: None )
		elif action == 'unpause':
			return self.parent.runAMICommand( 'queuePause' , queue=cmd.form.fields['queue'].values[0], 
				interface='sip/%s'%cmd.form.fields['member'].values[0], paused=False ).addCallback( lambda msg: None )
		elif action == 'remove':
			return self.parent.runAMICommand( 'queueRemove' , queue=cmd.form.fields['queue'].values[0],
				interface='sip/%s'%cmd.form.fields['member'].values[0] ).addCallback( lambda msg: None )

	def run_user_entity_tags( self, cmd ):
		for field in ("user",): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				raise error.StanzaError( 'not-acceptable' )
		tags = PermsDB.getUserEntityTags( cmd.form.fields["user"].values[0] )
		form = Form(formType='result')
		for tag in tags:
			form.addField(Field(var=tag,value=tag))
		cmd.form = form
		cmd.status = "completed"
		return [cmd.toElement()]

	def run_list_entities( self, cmd ):
		"""
		Retrieve all entityIds
		"""
		for field in ("entity_tag",): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				raise error.StanzaError( 'not-acceptable' )
		form = Form(formType='result')
		entityTag = cmd.form.fields["entity_tag"].values[0] 
		if entityTag == "extension":
			entities = ExtensDB.getExtens()
		elif entityTag == "voicemailbox":
			entities = MailboxDB.getMailboxes()
		elif entityTag == "conference":
			entities = ConfDB.getConfRooms()
		elif entityTag == "queue":
			entities = QueueDB.getQueues()
		elif entityTag == "parkinglot":
			entities = ["default"]
		else:
			raise error.StanzaError( 'not-acceptable' )
		for entity in entities:
			form.addField(Field(var=entity,value=entity))
		cmd.form = form
		cmd.status = 'completed'
		return [cmd.toElement()]

	def run_user_entities( self, cmd ):
		"""
		Retrieve all user related entity ids on an entity category
		"""
		for field in ("action", "user", "entity_tag",): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				raise error.StanzaError( 'not-acceptable' )
		if cmd.form.fields['action'].values[0] == 'get':
			form = Form(formType='result')
			for entity in PermsDB.getUserEntities( cmd.form.fields['user'].values[0], cmd.form.fields['entity_tag'].values[0] ):
				form.addField(Field(var=entity,value=entity))
			cmd.form = form
			cmd.status = 'completed'
			return [cmd.toElement()]
		elif cmd.form.fields['action'].values[0] == 'remove':
			PermsDB.removeUser(cmd.form.fields['user'].values[0], cmd.form.fields['entity_tag'].values[0])

	def run_entity_permissions( self, cmd ):
		"""
		Manages user/entity permissions
		"""
		for field in ('action', 'user', 'entity_tag','entity'): #Check required params
			if field not in cmd.form.fields or len(cmd.form.fields[field].values)==0 or len(cmd.form.fields[field].values[0])==0:
				raise error.StanzaError( 'not-acceptable' )
		# Get action here
		if cmd.form.fields["action"].values[0] == "get":
			form = Form(formType="result")
			entityPerms = PermsDB.getEntityUserPerms( cmd.form.fields['entity_tag'].values[0], 
				cmd.form.fields['entity'].values[0], cmd.form.fields['user'].values[0] )
			form.addField( Field(fieldType='list-multi',var="permissions",values=entityPerms ) )
			cmd.form  = form
			cmd.status = "completed"
			return [cmd.toElement()]

		if 'permissions' not in cmd.form.fields:
			raise error.StanzaError( 'not-acceptable' )
		if len(cmd.form.fields['permissions'].values)==0: # Remove command
			PermsDB.remove( cmd.form.fields['entity_tag'].values[0], cmd.form.fields['entity'].values[0], cmd.form.fields['user'].values[0] )
			return
		PermsDB.set( cmd.form.fields['entity_tag'].values[0], cmd.form.fields['entity'].values[0], 
			cmd.form.fields['user'].values[0], cmd.form.fields['permissions'].values )
		
