"""
Holds the actions related to extensions
"""

from twisted.python import log
from twisted.internet import reactor

from mbitrans import libmbit

EXTENSION_PERMS_FILE = '/var/lib/mbitrans/db/extenperms.pkl'

class ExtenPermsDb(object):
	"""
	Holds extensions related data
	"""
	config = {} # Will be populated here by tap module
	extensions = {} #{u'102': {u'dan': set([u'hangup', u'own'])}, u'101': {u'dan': set([u'hangup', u'own', u'select'])}} 
	statusNotifiers  = set() # list of functions to call on notify

	def __init__( self ):
		libmbit.loadData(EXTENSION_PERMS_FILE, self.extensions) #Load data from file on each startup

	def registerStatusNotifier( self, notifier ):
		"""External functions can use this method to register callback in case of presence db changes"""
		self.statusNotifiers.add( notifier )

	def unregisterStatusNotifier( self, notifier ):
		"""Unregister from receiving notifies on database changes"""
		self.statusNotifiers.discard( notifierMethod )

	def triggerStatusNotifies( self, user, action ):
		"""Registered functions will be called to inform about presence changes on specific entityName"""
		for notifier in self.statusNotifiers:
			reactor.callLater(self.config['gateway_delay'], notifier, user, action)

	def removeUserOwnExtension( self, user ):
		"""
		Method used to remove 'own' extension of a user in order to make space for the new one
		"""
		for extension in self.extensions.keys():
			if user in self.extensions[extension].keys() and 'own' in self.extensions[extension][user]:
				self.extensions[extension][user].remove('own')
				return #There can be only one such extesion, we found it so no need to waist time in looking further

	def set( self, extension, user, perms ):
		"""
		Add/Set extension and user
		@param perms: a set of permissions the user has on extension
		"""
		if 'own' in perms: #A new extension was set as own one for the user, so remove the old reference if we find one
			self.removeUserOwnExtension(user)
		if extension in self.extensions.keys():
			self.extensions[extension][user] = set(perms)
		else:
			self.extensions[extension] = { user: set(perms) }
		libmbit.fileData(EXTENSION_PERMS_FILE, self.extensions) #Save changes in realtime to disk
		self.triggerStatusNotifies(user, "set")
		

	def remove( self, extension, user ):
		"""
		Removes a user from extension data, extension if empty
		"""
		changed = False
		if extension in self.extensions.keys() and user in self.extensions[extension].keys():
			del self.extensions[extension][user]
			changed = True
		if extension in self.extensions.keys() and len(self.extensions[extension]) == 0: # Memory cleanup
			del self.extensions[extension]
			if not changed:
				changed = True
		if changed:
			libmbit.fileData(EXTENSION_PERMS_FILE, self.extensions) #Save changes in realtime to disk

	def removeUser( self, user ):
		"""
		Completely removes all rights related to an user
		"""
		changed = False
		for extension in self.extensions.keys():
			if user in self.extensions[extension].keys():
				del self.extensions[extension][user]
				if len(self.extensions[extension]) == 0: # Memory cleanup
					del self.extensions[extension]
				if not changed:
					changed = True
		if changed:
			libmbit.fileData(EXTENSION_PERMS_FILE, self.extensions) #Save changes in realtime to disk
			self.triggerStatusNotifies(user, "remove")

	def getExtenUserPerms( self, exten, user ):
		try:
			return self.extensions[exten][user]
		except KeyError:
			return []

	def getExtenUsers( self, exten ):
		try:
			return self.extensions[exten].keys()
		except KeyError:
			return []

	def getUserExtens( self, user ):
		"""
		@returns extensions list attached to an user
		"""
		return [ ext for ext in self.extensions if user in self.extensions[ext] ]
			

	def getUserExtensWithPerms( self, user ):
		"""
		Returns extensions together with their permissions for an user
		"""
		userExtens = {}
		exts = [ ext for ext in self.extensions.keys() if user in self.extensions[ext].keys() ]
		for ext in exts:
			userExtens[ext] = self.extensions[ext][user]
		return userExtens

	def getUserExtensWithPerm( self, user, perm ):
		return [exten for exten in self.extensions if (user in self.extensions[exten].keys() and perm in self.extensions[exten][user])]

	def getExtensions( self ):
		return self.extensions


	def getUsersWithExtenPerm( self, exten, perm ):
		"""
		@return all users having specific permission on extension
		"""
		if exten not in self.extensions.keys():
			return []
		return [ user for user in self.extensions[exten] if perm in self.extensions[exten][user] ]

	def getAllUsers( self ):
		"""
		Returns a set of all users related to extensions
		"""
		users = set();
		for extension in self.extensions.keys():
			users.update(self.extensions[extension])
		return users

ExtenPermsDB = ExtenPermsDb()
