"""
Generic permissions management
"""

from twisted.python import log
from twisted.internet import reactor

from mbitrans import libmbit

PERMS_FILE = '/var/lib/mbitrans/db/perms.pkl'

class PermsDb(object):
	"""
	Processor for user/entity permissions
	"""

	config = {} # Will be populated here by tap module
	perms = {} #{"entityTag":{u'entityId': {u'dan': set([u'hangup', u'own'])}}
	statusNotifiers  = {} #{"entityTag":set()} # list of functions to call on notify

	def __init__( self ):
		libmbit.loadData(PERMS_FILE, self.perms) #Load data from file on each startup

	def registerStatusNotifier( self, entityTag, notifier ):
		"""External functions can use this method to register callback in case of perms db changes"""
		if entityTag not in self.statusNotifiers:
			self.statusNotifiers[entityTag] = set()
		self.statusNotifiers[entityTag].add( notifier )

	def unregisterStatusNotifier( self, entityTag, notifier ):
		"""Unregister from receiving notifies on database changes"""
		if entityTag in self.statusNotifiers:
			self.statusNotifiers.discard( notifierMethod )

	def triggerStatusNotifies( self, entityTag, user, action ):
		"""Registered functions will be called to inform about permissions change for specific user"""
		notifiers = set()
		tags = (entityTag, "*",)
		for tag in tags:
			if tag in self.statusNotifiers.keys():
				notifiers.update(self.statusNotifiers[tag])
		for notifier in notifiers:
			reactor.callLater(self.config['gateway_delay'], notifier, entityTag, user, action)

	def removeUserOwnExtension( self, user ):
		"""
		Method used to remove 'own' extension of a user in order to make space for the new one
		"""
		if "extension" not in self.perms.keys():
			return # Did not set any extension perms, no own defined so we are OK
		for extension in self.perms['extension'].keys():
			if user in self.perms['extension'][extension].keys() and 'own' in self.perms["extension"][extension][user]:
				self.perms['extension'][extension][user].remove('own')
				return #There can be only one such extesion, we found it so no need to waist time in looking further

	def set( self, entityTag, entityId, user, perms ):
		"""
		Add/Set entity perms for an user
		@param perms: a set of permissions the user has on entity 
		"""
		if entityTag=="extension" and "own" in perms: #A new extension was set as own one for the user, so remove the old reference if we find one
			self.removeUserOwnExtension(user)
		if entityTag not in self.perms.keys():
			self.perms[entityTag] = {}
		if entityId not in self.perms[entityTag]:
			self.perms[entityTag][entityId] = {}
		self.perms[entityTag][entityId][user] = set(perms)
		libmbit.fileData(PERMS_FILE, self.perms) #Save changes in realtime to disk
		self.triggerStatusNotifies( entityTag, user, "set")
		

	def remove( self, entityTag, entityId, user ):
		"""
		Removes a user from entity database, entityId if empty
		"""
		changed = False
		if entityTag in self.perms.keys() and entityId in self.perms[entityTag].keys() and user in self.perms[entityTag][entityId].keys():
			del self.perms[entityTag][entityId][user]
			if len(self.perms[entityTag][entityId]) == 0:
				del self.perms[entityTag][entityId]
			changed = True
		if changed:
			libmbit.fileData(PERMS_FILE, self.perms) #Save changes in realtime to disk

	def removeUser( self, user, entityTag ):
		"""
		Completely removes all perms related to an user on an entityTag
		"""
		changed = False
		if entityTag not in self.perms:
			return
		for entityId in self.perms[entityTag].keys():
			if user in self.perms[entityTag][entityId].keys():
				del self.perms[entityTag][entityId][user]
				if len(self.perms[entityTag][entityId]) == 0: # Memory cleanup
					del self.perms[entityTag][entityId]
				changed = True
		if changed:
			libmbit.fileData(PERMS_FILE, self.perms) #Save changes in realtime to disk
			self.triggerStatusNotifies( entityTag, user, "remove")

	def getEntityUserPerms( self, entityTag, entityId, user ):
		try:
			return self.perms[entityTag][entityId][user]
		except KeyError:
			return set()

	def getEntityUsers( self, entityTag, entityId ):
		try:
			return self.perms[entityTag][entityId].keys()
		except KeyError:
			return []

	def getUserPerms( self, user ):
		"""
		Returns all userPerms  related to an user
		"""
		perms = {}
		for entityTag in self.perms.keys():
			perms[entityTag] = {}
			for entityId in self.perms[entityTag].keys():
				if user in self.perms[entityTag][entityId].keys():
					perms[entityTag][entityId] = self.perms[entityTag][entityId][user]
		return perms

	def getUserEntities( self, user, entityTag ):
		if entityTag in self.perms:
			return [ entityId for entityId in self.perms[entityTag] if user in self.perms[entityTag][entityId].keys() ]
		return []
			
	def getUserEntitiesHavingPerm( self, entityTag, user, perm ):
		if not entityTag in self.perms:
			return []
		return [ entityId for entityId in self.perms[entityTag] 
			if user in self.perms[entityTag][entityId].keys() and perm in self.perms[entityTag][entityId][user] ]

	def getUserEntitiesHavingOneOfPerms( self, entityTag, user, perms ):
		entities = set()
		for perm in perms:
			entities.update( self.getUserEntitiesHavingPerm( entityTag, user, perm ) )
		return entities

	def getUserEntityTags( self, user ):
		tags = set()
		for entityTag in self.perms.keys():
			for entityId in self.perms[entityTag].keys():
				if user in self.perms[entityTag][entityId]:
					tags.add( entityTag )
		return tags

	def getEntityUsersHavingPerm( self, entityTag, entityId, perm ):
		if entityTag not in self.perms.keys() or entityId not in self.perms[entityTag].keys():
			return []
		return [ user for user in self.perms[entityTag][entityId] if perm in self.perms[entityTag][entityId][user] ]

	def getEntityUsersHavingOneOfPerms( self, entityTag, entityId, perms ):
		"""
		@param perms: a set of permissions which acts as or condition
		"""
		users = set()
		for perm in perms:
			users.update( self.getEntityUsersHavingPerm( entityTag, entityId, perm ) )
		return users

	def getAllUsers( self ):
		users = set();
		for entityTag in self.perms.keys():
			for entityId in self.perms[entityTag].keys():
				users.update(self.perms[entityTag][entityId].keys())
		return users

	def getPerms( self ):
		return dict(self.perms)

PermsDB = PermsDb()
