from wokkel.generic import Request
from wokkel import pubsub, disco
from twisted.words.xish.domish import Element
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.error import StanzaError

from twisted.python import log
from twisted.internet import defer
from time import time

from mbitrans.presencedb import PresenceDB
from mbitrans.extenpermsdb import ExtenPermsDB
from mbitrans.extensdb import ExtensDB
from mbitrans.callsdb import CallsDB
from mbitrans.parkdb import ParkDB
from mbitrans.confdb import ConfDB
from mbitrans.mailboxdb import MailboxDB
from mbitrans.queuedb import QueueDB
from mbitrans.permsdb import PermsDB
from mbitrans.settingsdb import SettingsDB

import sys
NS_ACALLS = "urn:xmpp:mbit:acalls"
XPATH_ACALLS_GET = "/iq[@type='get']/query[@xmlns='%s']" % NS_ACALLS


class PubServHandler( pubsub.PubSubService ):
	"""
	I am in charge of monitoring live calls and publishing them towards contacts in the entity's roster
	"""
	pubSubFeatures = ['presence-subscribe']

	nodesHandled = ( 'extension_status', 'extensions_status', 'active_calls', 'parked_calls', 'mailboxes', 'queue_calls', 'queue_agents', 'queue_stats' )

	def __init__(self, resource=None):
		self.resource = resource
		self.discoIdentity = {'category': 'pubsub',
				'type': 'service',
				'name': 'AMI Publish Service'}

	def getDiscoInfo(self, requestor, target, nodeIdentifier=''):
		"""
		Used mostly to discover the node configuration
		"""
		if nodeIdentifier in self.nodesHandled: #Right contact for right entity, return node config as leaf
			return defer.succeed([disco.DiscoIdentity(category='pubsub', idType='leaf')])
		return defer.succeed([]) #Error if we return none, so empty list will do

	def getDiscoItems(self, requestor, target, nodeIdentifier=''):
		"""
		Called on disco#items
		"""
		return defer.succeed([])

	def connectionMade(self):
		self.xmlstream.addObserver(pubsub.PUBSUB_REQUEST, self.handleRequest)
		self.registerForNotifies()


	def registerForNotifies( self ):
		PresenceDB.registerStatusNotifier( self.onUserPresenceChange ) #On presence change will be pushed out all 
		ExtensDB.registerStatusNotifier( self.onExtenStatusChange )
		SettingsDB.registerStatusNotifier( self.onSettingsChange )
		CallsDB.registerStatusNotifier( self.onCallStatusChange )
		MailboxDB.registerStatusNotifier( self.onMboxChange )
		ParkDB.registerStatusNotifier( self.onParkedChange )
		ConfDB.registerStatusNotifier( self.onConfChange )
		QueueDB.registerStatusNotifier( 'queued', self.onQueueCallsChange )
		QueueDB.registerStatusNotifier( 'agents', self.onQueueAgentsChange )
		QueueDB.registerStatusNotifier( 'stats', self.onQueueStatsChange )

	def onUserPresenceChange( self, user ):
		if PresenceDB.getUserPresence(user)[0] != "available":
			return #User must be online to receive notifications
		self.pubUserPerms( user )
		self.pubUserExtensStatus( user )
		self.pubUserCalls( user )
		self.pubUserMailboxes( user )
		self.pubParkedCalls( user )
		self.pubUserConferences( user )
		self.pubUserQueuesCalls( user )
		self.pubUserQueuesAgents( user )
		self.pubUserQueuesStats( user )

	def getCategPermsItem( self, entityTag, entitiesPerms ):
		"""
		@returns: permissions item  for specific user
		"""
		item = Element((None, 'item'))
		categElmnt = Element((None, "category"), attribs={"tag":entityTag})
		for entityId in entitiesPerms:
			entityElmnt = Element((None, 'element'), attribs={"id":entityId})
			for perm in entitiesPerms[entityId]:
				entityElmnt.addElement('permission', content=perm)
			categElmnt.addChild( entityElmnt )
		item.addChild(categElmnt)
		return item

	def pubUserPerms( self, user ):
		"""
		Publish user permissions when he comes online (probe will enforce also on permissions set to be resent
		"""
		items = []
		userPerms = PermsDB.getUserPerms( user )
		for entityTag in userPerms:
			items.append( self.getCategPermsItem( entityTag, userPerms[entityTag] ) )
		return self.notifyPublish( self.parent.myJID(), 'category_permissions', [[JID(user), None, items]] )

	def getExtenStatusItem( self, extension ):
		"""
		@returns: extension item
		"""
		extenItem = Element((None, 'item'),attribs={"id":extension})
		extenElmnt = Element((None, 'extension'), attribs={'id':extension, 'status':ExtensDB.getExtenStatus(extension), 
			'name': SettingsDB.getEntityProperty(extension, 'name', 'unknown')})
		extenItem.addChild( extenElmnt )
		return extenItem

	def pubUserExtensStatus( self, user ):
		"""
		Publish user extensions, eg when user comes online
		"""
		items = []
		for exten in PermsDB.getUserEntitiesHavingPerm( "extension", user, "view" ):
			items.append( self.getExtenStatusItem( exten ) )
		return self.notifyPublish( self.parent.myJID(), 'extensions_status', [[JID(user), None, items]] )
		
	def onExtenStatusChange( self, exten ):
		"""
		Publish extension status on status change
		"""
		notifications = []
		for user in PermsDB.getEntityUsersHavingPerm( "extension", exten, "view" ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, [self.getExtenStatusItem(exten)]] )
		return self.notifyPublish( self.parent.myJID(), 'extension_status', notifications )

	def onSettingsChange( self, entity, key ):
		"""
		When settings will be changed for entity we will be informed. Useful in case of name changes
		"""
		if key == 'name':
			self.onExtenStatusChange( entity )

	def getCallItems( self, entityName ):
		items = []
		extenItem = Element((None, 'item'), attribs={'id':self.parent.entityJID(entityName).userhost()})
		extenElmnt = Element((None, 'extension'), attribs={'jid':self.parent.entityJID(entityName).userhost()})
		extenItem.addChild( extenElmnt )
		for callDetails in CallsDB.getExtensionCalls( entityName ):
			#[callId, direction, from, to, chanStatus, setupTime, connectTime, context, bChanId]
			callDur = ''
			if callDetails[6]:
				callDur = str(int(time()-callDetails[6])) #The Element does not want to send integer out
			attribs = {'direction':callDetails[1], 'chanid': callDetails[0], 'bChanId': callDetails[8],
				'calling': callDetails[2], 'called': callDetails[3],
				'status': callDetails[4], 'duration': callDur}
			extenElmnt.addChild( Element((None, 'call'), attribs = attribs ) )
		items.append( extenItem )
		return items

	def pubUserCalls( self, user ):
		"""
		Publish user calls
		"""
		items = []
		for exten in PermsDB.getUserEntitiesHavingPerm( "extension", user, "view" ):
			items += self.getCallItems( exten )
		return self.notifyPublish( self.parent.myJID(), 'active_calls', [[JID(user), None, items]] )

	def onCallStatusChange( self, entityName ):
		"""
		This will be called by CallsDB when there will be a changes on calls of an entity
		@param entityName: the name of entity as index for the calls
		@type entityName: C{string}
		"""
		notifications = []
		for user in PermsDB.getEntityUsersHavingPerm( "extension", entityName, "view" ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getCallItems( entityName )] )
		return self.notifyPublish( self.parent.myJID(), 'active_calls', notifications )

	def getParkedItems( self ):
		"""
		Responsible of formating parked calls data as useful information in xmpp world
		"""
		items = []
		for parkedCall in ParkDB.getParkedCalls():
			#['SIP/101-00000017', 'default', '701', '101', '<unknown>', '45', 1356532285.532025]
			item = Element((None, 'item'), attribs={'id':parkedCall[0]})
			timeout = str( int(parkedCall[5]) - int(time()-parkedCall[6]) ) #The Element does not want to send integer out
			attribs = {'chanid':parkedCall[0], 'parklot':parkedCall[1], 'parkextension':parkedCall[2], 
				'calling': parkedCall[3], 'called': parkedCall[4],
				'timeout': timeout, 'extension': parkedCall[7]}
			item.addChild( Element((None, 'call'), attribs = attribs ) )
			items.append( item )
		return items

	def pubParkedCalls( self, user ):
		items = self.getParkedItems()
		return self.notifyPublish( self.parent.myJID(), 'parked_calls', [[JID(user), None, items]] )

	def onParkedChange( self ):
		notifications = []
		for user in PermsDB.getEntityUsersHavingPerm( "parkinglot", "default", "view" ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getParkedItems()] )
		return self.notifyPublish( self.parent.myJID(), 'parked_calls', notifications )


	def getConfRoomItems( self, confRoom ):
		"""
		Responsible of formating conference calls data as useful information in xmpp world
		"""
		items = []
		item = Element((None, 'item'), attribs={'id':confRoom})
		confroomElmnt = Element((None, 'confroom'), attribs={'id':confRoom})
		item.addChild( confroomElmnt )
		for confCall in ConfDB.getConfRoom( confRoom ):
			# ['SIP/101-00000003', '101', '1', 1356606071.6245141]
			calldur = str( int(time()-confCall[3]) )
			attribs = {'chanid':confCall[0], 'caller':confCall[1], 'usernr':str(confCall[2]), 'duration':calldur}
			confroomElmnt.addChild( Element((None, 'confcall'), attribs = attribs ) )
		items.append( item )
		return items

	def pubUserConferences( self, user ):
		"""
		Publish user calls
		"""
		items = []
		for entity in PermsDB.getUserEntitiesHavingOneOfPerms( "conference", user, ("administer", "view") ):
			items += self.getConfRoomItems( entity )
		return self.notifyPublish( self.parent.myJID(), 'conferences_calls', [[JID(user), None, items]] )

	def onConfChange( self, confRoom ):
		notifications = []
		xmppUsers = PermsDB.getEntityUsersHavingOneOfPerms( "conference", confRoom, ("administer", "view") )
		for user in xmppUsers:
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getConfRoomItems(confRoom)] )
		return self.notifyPublish( self.parent.myJID(), 'conference_calls', notifications )

	def getMailboxItems( self, vmbox ):
		items = []
		item = Element((None, 'item'), attribs={'id':vmbox})
		mailbox = MailboxDB.getMailbox(vmbox)
		if mailbox and len(mailbox)==4: # Should be in the form of [vmid, newmsg, oldmsg, urgmsg]
			mailboxElmnt = Element((None, 'voicemailbox'), attribs={'id':mailbox[0],'new':mailbox[1], 'old':mailbox[2], 'urgent':mailbox[3]})
			item.addChild( mailboxElmnt )
		items.append( item )
		return items

	def pubUserMailboxes( self, user ):
		items = []
		for mbox in PermsDB.getUserEntitiesHavingPerm( "voicemailbox", user, "view" ):
			items += self.getMailboxItems(mbox)
		return self.notifyPublish( self.parent.myJID(), 'voicemailboxes', [[JID(user), None, items]] )


	def onMboxChange( self, vmbox ):
		notifications = []
		for user in PermsDB.getEntityUsersHavingPerm( "voicemailbox", vmbox, "view" ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getMailboxItems(vmbox)] )
		return self.notifyPublish( self.parent.myJID(), 'voicemailbox', notifications )


	def getQueueCallItems( self, queue ):
		items = []
		item = Element((None, 'item'), attribs={'id':queue})
		queueElmnt = Element((None, 'queue'), attribs={'id':queue})
		item.addChild( queueElmnt )
		for call in QueueDB.getQueueCalls( queue ):
			attribs = {'position':call[1], 'caller':call[2], 'chanid':call[0]}
			queueElmnt.addChild( Element((None, 'call'), attribs = attribs ) )
		items.append( item )
		return items

	def pubUserQueuesCalls( self, user ):
		items = []
		for entity in PermsDB.getUserEntitiesHavingOneOfPerms( "queue", user, ("view", "admin") ):
			items += self.getQueueCallItems(entity)
		return self.notifyPublish( self.parent.myJID(), 'queues_calls', [[JID(user), None, items]] )

	def onQueueCallsChange( self, queue ):
		notifications = []
		for user in PermsDB.getEntityUsersHavingOneOfPerms( "queue", queue, ("view", "admin") ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getQueueCallItems(queue)] )
		return self.notifyPublish( self.parent.myJID(), 'queue_calls', notifications )

	def getQueueAgentItems( self, queue ):
		items = []
		item = Element((None, 'item'), attribs={'id':queue})
		queueElmnt = Element((None, 'queue'), attribs={'id':queue})
		item.addChild( queueElmnt )
		for agent in QueueDB.getQueueAgents( queue ):
			attribs = {'id':agent[0], 'membership':agent[1], 'status': agent[2], 'location': agent[3],
				'callstaken':agent[4], 'paused':agent[6], 'penalty':agent[7]}
			queueElmnt.addChild( Element((None, 'agent'), attribs = attribs ) )
		items.append( item )
		return items

	def pubUserQueuesAgents( self, user ):
		items = []
		for entity in PermsDB.getUserEntitiesHavingOneOfPerms( "queue", user, ("view", "admin") ):
			items += self.getQueueAgentItems( entity )
		return self.notifyPublish( self.parent.myJID(), 'queues_agents', [[JID(user), None, items]] )

	def onQueueAgentsChange( self, queue ):
		notifications = []
		for user in PermsDB.getEntityUsersHavingOneOfPerms( "queue", queue, ("view", "admin") ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getQueueAgentItems(queue)] )
		return self.notifyPublish( self.parent.myJID(), 'queue_agents', notifications )


	def getQueueStatsItems( self, queue ):
		items = []
		item = Element((None, 'item'), attribs={'id':queue})
		queueElmnt = Element((None, 'queue'), attribs={'id':queue})
		item.addChild( queueElmnt )
		queueElmnt.addChild( Element((None, 'stats'), attribs = QueueDB.getQueueStats( queue )) )
		items.append( item )
		return items

	def pubUserQueuesStats( self, user ):
		items = []
		for entity in PermsDB.getUserEntitiesHavingPerm( "queue", user, "admin" ):
			items += self.getQueueStatsItems( entity )
		return self.notifyPublish( self.parent.myJID(), 'queues_stats', [[JID(user), None, items]] )

	def onQueueStatsChange( self, queue ):
		notifications = []
		for user in PermsDB.getEntityUsersHavingPerm( "queue", queue, "admin" ):
			if PresenceDB.getUserPresence(user)[0] != "available":
				continue # Do nothing if user is not online
			notifications.append( [JID(user), None, self.getQueueStatsItems(queue)] )
		return self.notifyPublish( self.parent.myJID(), 'queue_stats', notifications )
		

